@@ -8,17 +8,16 @@
 -- Again, from the top!
 
 local t_insert = table.insert;
-local jid = require "util.jid";
-local st = require "util.stanza";
+local jid = require "prosody.util.jid";
+local st = require "prosody.util.stanza";
 local url = require "socket.url";
-local dm = require "core.storagemanager".olddm;
-local jwt = require "util.jwt";
-local errors = require "util.error";
-local dataform = require "util.dataforms".new;
-local urlencode = require "util.http".urlencode;
-local dt = require "util.datetime";
-local hi = require "util.human.units";
-local cache = require "util.cache";
+local dm = require "prosody.core.storagemanager".olddm;
+local errors = require "prosody.util.error";
+local dataform = require "prosody.util.dataforms".new;
+local urlencode = require "prosody.util.http".urlencode;
+local dt = require "prosody.util.datetime";
+local hi = require "prosody.util.human.units";
+local cache = require "prosody.util.cache";
 local lfs = require "lfs";
 
 local unknown = math.abs(0/0);
@@ -35,17 +34,21 @@ local uploads = module:open_store("uploads", "archive");
 local persist_stats = module:open_store("upload_stats", "map");
 -- id, <request>, time, owner
 
-local secret = module:get_option_string(module.name.."_secret", require"util.id".long());
+local secret = module:get_option_string(module.name.."_secret", require"prosody.util.id".long());
 local external_base_url = module:get_option_string(module.name .. "_base_url");
-local file_size_limit = module:get_option_number(module.name .. "_size_limit", 10 * 1024 * 1024); -- 10 MB
+local file_size_limit = module:get_option_integer(module.name .. "_size_limit", 10 * 1024 * 1024, 0); -- 10 MB
 local file_types = module:get_option_set(module.name .. "_allowed_file_types", {});
 local safe_types = module:get_option_set(module.name .. "_safe_file_types", {"image/*","video/*","audio/*","text/plain"});
-local expiry = module:get_option_number(module.name .. "_expires_after", 7 * 86400);
-local daily_quota = module:get_option_number(module.name .. "_daily_quota", file_size_limit*10); -- 100 MB / day
-local total_storage_limit = module:get_option_number(module.name.."_global_quota", unlimited);
+local expiry = module:get_option_period(module.name .. "_expires_after", "1w");
+local daily_quota = module:get_option_integer(module.name .. "_daily_quota", file_size_limit*10, 0); -- 100 MB / day
+local total_storage_limit = module:get_option_integer(module.name.."_global_quota", unlimited, 0);
+
+local create_jwt, verify_jwt = require"prosody.util.jwt".init("HS256", secret, secret, { default_ttl = 600 });
 
 local access = module:get_option_set(module.name .. "_access", {});
 
+module:default_permission("prosody:registered", ":upload");
+
 if not external_base_url then
 	module:depends("http");
 end
@@ -59,12 +62,33 @@ local upload_errors = errors.init(module.name, namespace, {
 	access = { type = "auth"; condition = "forbidden" };
 	filename = { type = "modify"; condition = "bad-request"; text = "Invalid filename" };
 	filetype = { type = "modify"; condition = "not-acceptable"; text = "File type not allowed" };
-	filesize = { type = "modify"; condition = "not-acceptable"; text = "File too large";
-		extra = {tag = st.stanza("file-too-large", {xmlns = namespace}):tag("max-file-size"):text(tostring(file_size_limit)) };
+	filesize = {
+		code = 413;
+		type = "modify";
+		condition = "not-acceptable";
+		text = "File too large";
+		extra = {
+			tag = st.stanza("file-too-large", { xmlns = namespace }):tag("max-file-size"):text(tostring(file_size_limit));
+		};
 	};
 	filesizefmt = { type = "modify"; condition = "bad-request"; text = "File size must be positive integer"; };
 	quota = { type = "wait"; condition = "resource-constraint"; text = "Daily quota reached"; };
 	outofdisk = { type = "wait"; condition = "resource-constraint"; text = "Server global storage quota reached" };
+	authzmalformed = {
+		code = 401;
+		type = "auth";
+		condition = "not-authorized";
+		text = "Missing or malformed Authorization header";
+	};
+	unauthz = { code = 403; type = "auth"; condition = "forbidden"; text = "Unauthorized or invalid token" };
+	invalidslot = {
+		code = 400;
+		type = "modify";
+		condition = "bad-request";
+		text = "Invalid upload slot, must not contain '/'";
+	};
+	alreadycompleted = { code = 409; type = "cancel"; condition = "conflict"; text = "Upload already completed" };
+	writefail = { code = 500; type = "wait"; condition = "internal-server-error" }
 });
 
 local upload_cache = cache.new(1024);
@@ -135,7 +159,7 @@ end
 
 function may_upload(uploader, filename, filesize, filetype) -- > boolean, error
 	local uploader_host = jid.host(uploader);
-	if not ((access:empty() and prosody.hosts[uploader_host]) or access:contains(uploader) or access:contains(uploader_host)) then
+	if not (module:may(":upload", uploader) or access:contains(uploader) or access:contains(uploader_host)) then
 		return false, upload_errors.new("access");
 	end
 
@@ -169,16 +193,13 @@ function may_upload(uploader, filename, filesize, filetype) -- > boolean, error
 end
 
 function get_authz(slot, uploader, filename, filesize, filetype)
-local now = os.time();
-	return jwt.sign(secret, {
+	return create_jwt({
 		-- token properties
 		sub = uploader;
-		iat = now;
-		exp = now+300;
 
 		-- slot properties
 		slot = slot;
-		expires = expiry >= 0 and (now+expiry) or nil;
+		expires = expiry < math.huge and (os.time()+expiry) or nil;
 		-- file properties
 		filename = filename;
 		filesize = filesize;
@@ -224,6 +245,7 @@ function handle_slot_request(event)
 	end
 
 	total_storage_usage = total_storage_usage + filesize;
+	persist_stats:set(nil, "total", total_storage_usage);
 	module:log("debug", "Total storage usage: %s / %s", B(total_storage_usage), B(total_storage_limit));
 
 	local cached_quota = quota_cache:get(uploader);
@@ -249,32 +271,34 @@ end
 
 function handle_upload(event, path) -- PUT /upload/:slot
 	local request = event.request;
-	local authz = request.headers.authorization;
-	if authz then
-		authz = authz:match("^Bearer (.*)")
-	end
-	if not authz then
-		module:log("debug", "Missing or malformed Authorization header");
-		event.response.headers.www_authenticate = "Bearer";
-		return 401;
-	end
-	local authed, upload_info = jwt.verify(secret, authz);
-	if not (authed and type(upload_info) == "table" and type(upload_info.exp) == "number") then
-		module:log("debug", "Unauthorized or invalid token: %s, %q", authed, upload_info);
-		return 401;
-	end
-	if not request.body_sink and upload_info.exp < os.time() then
-		module:log("debug", "Authorization token expired on %s", dt.datetime(upload_info.exp));
-		return 410;
-	end
-	if not path or upload_info.slot ~= path:match("^[^/]+") then
-		module:log("debug", "Invalid upload slot: %q, path: %q", upload_info.slot, path);
-		return 400;
-	end
-	if request.headers.content_length and tonumber(request.headers.content_length) ~= upload_info.filesize then
-		return 413;
-		-- Note: We don't know the size if the upload is streamed in chunked encoding,
-		-- so we also check the final file size on completion.
+	local upload_info = request.http_file_share_upload_info;
+
+	if not upload_info then -- Initial handling of request
+		local authz = request.headers.authorization;
+		if authz then
+			authz = authz:match("^Bearer (.*)")
+		end
+		if not authz then
+			module:log("debug", "Missing or malformed Authorization header");
+			event.response.headers.www_authenticate = "Bearer";
+			return upload_errors.new("authzmalformed", { request = request });
+		end
+		local authed, authed_upload_info = verify_jwt(authz);
+		if not authed then
+			module:log("debug", "Unauthorized or invalid token: %s, %q", authz, authed_upload_info);
+			return upload_errors.new("unauthz", { request = request; wrapped_error = authed_upload_info });
+		end
+		if not path or authed_upload_info.slot ~= path:match("^[^/]+") then
+			module:log("debug", "Invalid upload slot: %q, path: %q", authed_upload_info.slot, path);
+			return upload_errors.new("unauthz", { request = request });
+		end
+		if request.headers.content_length and tonumber(request.headers.content_length) ~= authed_upload_info.filesize then
+			return upload_errors.new("filesize", { request = request });
+			-- Note: We don't know the size if the upload is streamed in chunked encoding,
+			-- so we also check the final file size on completion.
+		end
+		upload_info = authed_upload_info;
+		request.http_file_share_upload_info = upload_info;
 	end
 
 	local filename = get_filename(upload_info.slot, true);
@@ -285,7 +309,7 @@ function handle_upload(event, path) -- PUT /upload/:slot
 		local f = io.open(filename, "r");
 		if f then
 			f:close();
-			return 409;
+			return upload_errors.new("alreadycompleted", { request = request });
 		end
 	end
 
@@ -294,7 +318,7 @@ function handle_upload(event, path) -- PUT /upload/:slot
 		local fh, err = io.open(filename.."~", "w");
 		if not fh then
 			module:log("error", "Could not open file for writing: %s", err);
-			return 500;
+			return upload_errors.new("writefail", { request = request; wrapped_error = err });
 		end
 		function event.response:on_destroy() -- luacheck: ignore 212/self
 			-- Clean up incomplete upload
@@ -327,7 +351,7 @@ function handle_upload(event, path) -- PUT /upload/:slot
 		local uploaded, err = errors.coerce(request.body_sink:close());
 		if final_size ~= upload_info.filesize then
 			-- Could be too short as well, but we say the same thing
-			uploaded, err = false, 413;
+			uploaded, err = false, upload_errors.new("filesize", { request = request });
 		end
 		if uploaded then
 			module:log("debug", "Upload of %q completed, %s", filename, B(final_size));
@@ -405,18 +429,22 @@ function handle_download(event, path) -- GET /uploads/:slot+filename
 	local request_range = request.headers.range;
 	local response_range;
 	if request_range then
+		local last_byte = string.format("%d", tonumber(filesize) - 1);
 		local range_start, range_end = request_range:match("^bytes=(%d+)%-(%d*)$")
 		-- Only support resumption, ie ranges from somewhere in the middle until the end of the file.
-		if (range_start and range_start ~= "0") and (range_end == "" or range_end == filesize) then
+		if (range_start and range_start ~= "0") and (range_end == "" or range_end == last_byte) then
 			local pos, size = tonumber(range_start), tonumber(filesize);
 			local new_pos = pos < size and handle:seek("set", pos);
 			if new_pos and new_pos < size then
-				response_range = "bytes "..range_start.."-"..filesize.."/"..filesize;
+				response_range = "bytes "..range_start.."-"..last_byte.."/"..filesize;
 				filesize = string.format("%d", size-pos);
 			else
 				handle:close();
 				return 416;
 			end
+		else
+			handle:close();
+			return 416;
 		end
 	end
 
@@ -450,11 +478,11 @@ function handle_download(event, path) -- GET /uploads/:slot+filename
 	return response:send_file(handle);
 end
 
-if expiry >= 0 and not external_base_url then
+if expiry < math.huge and not external_base_url then
 	-- TODO HTTP DELETE to the external endpoint?
-	local array = require "util.array";
-	local async = require "util.async";
-	local ENOENT = require "util.pposix".ENOENT;
+	local array = require "prosody.util.array";
+	local async = require "prosody.util.async";
+	local ENOENT = require "prosody.util.pposix".ENOENT;
 
 	local function sleep(t)
 		local wait, done = async.waiter();
@@ -578,6 +606,7 @@ if not external_base_url then
 module:provides("http", {
 		streaming_uploads = true;
 		cors = {
+			enabled = true;
 			credentials = true;
 			headers = {
 				Authorization = true;
