@@ -1,3 +1,5 @@
+local m_type = math.type;
+
 local function unescape_token(escaped_token)
 	local unescaped = escaped_token:gsub("~1", "/"):gsub("~0", "~")
 	return unescaped
@@ -15,12 +17,12 @@ local function resolve_json_pointer(ref, path)
 
 		if type(idx) == "string" then
 			new_ref = ref[token]
-		elseif math.type(idx) == "integer" then
+		elseif m_type(idx) == "integer" then
 			local i = tonumber(token)
 			if token == "-" then
 				i = #ref + 1
 			end
-			new_ref = ref[i]
+			new_ref = ref[i + 1]
 		else
 			return nil, "invalid-table"
 		end
