@@ -3,6 +3,7 @@ local data_path = '/var/prosody';
 local vhost = {
 	"accounts",
 	"account_details",
+	"account_roles",
 	"roster",
 	"vcard",
 	"private",
@@ -12,18 +13,27 @@ local vhost = {
 	"offline-archive",
 	"pubsub_nodes-pubsub",
 	"pep-pubsub",
+	"cron",
+	"smacks_h",
 }
 local muc = {
 	"persistent",
 	"config",
 	"state",
 	"muc_log-archive",
+	"cron",
 };
+local upload = {
+	"uploads-archive",
+	"upload_stats",
+	"cron",
+}
 
 input {
 	hosts = {
 		["example.com"] = vhost;
 		["conference.example.com"] = muc;
+		["share.example.com"] = upload;
 	};
 	type = "internal";
 	path = data_path;
